open Binsec
open Libsse
open Types

include Cli.Make (struct
  let name = "Cram test generator"
  let shortname = "cram-test"
end)

module Output = Builder.String_option (struct
  let name = "output"
  let doc = "Path of the output file"
end)

let () =
  Exec.register_plugin
    (module struct
      let name = "tuto"

      type Ast.Instr.t +=
        | Gets of Ast.Expr.t Ast.loc * Ast.Expr.t Ast.loc
        | Puts of Ast.Expr.t Ast.loc
        | Exit

      let grammar_extension =
        [
          Dyp.Add_rules
            [
              ( ( "fallthrough",
                  [
                    Dyp.Regexp (RE_String "gets");
                    Dyp.Regexp (RE_Char '(');
                    Dyp.Non_ter ("expr", No_priority);
                    Dyp.Regexp (RE_Char ',');
                    Dyp.Non_ter ("expr", No_priority);
                    Dyp.Regexp (RE_Char ')');
                  ],
                  "default_priority",
                  [] ),
                fun _ -> function
                  | [
                      _;
                      _;
                      Libparser.Syntax.Expr ptr;
                      _;
                      Libparser.Syntax.Expr size;
                      _;
                    ] ->
                      (Libparser.Syntax.Instr (Gets (ptr, size)), [])
                  | _ -> assert false );
              ( ( "fallthrough",
                  [
                    Dyp.Regexp (RE_String "puts");
                    Dyp.Regexp (RE_Char '(');
                    Dyp.Non_ter ("expr", No_priority);
                    Dyp.Regexp (RE_Char ')');
                  ],
                  "default_priority",
                  [] ),
                fun _ -> function
                  | [ _; _; Libparser.Syntax.Expr ptr; _ ] ->
                      (Libparser.Syntax.Instr (Puts ptr), [])
                  | _ -> assert false );
              ( ( "fallthrough",
                  [
                    Dyp.Regexp (RE_String "exit");
                    Dyp.Regexp (RE_Char '(');
                    Dyp.Regexp (RE_Char ')');
                  ],
                  "default_priority",
                  [] ),
                fun _ -> function
                  | [ _; _; _ ] -> (Libparser.Syntax.Instr Exit, [])
                  | _ -> assert false );
            ];
        ]

      let instruction_printer = None
      let declaration_printer = None

      let extension :
          type a b.
          (module EXPLORATION_STATISTICS) ->
          (module Path.S with type t = a) ->
          (module STATE with type t = b) ->
          (module Exec.EXTENSION with type path = a and type state = b) option =
       fun _stats path state ->
        if is_enabled () then
          Some
            (module struct
              module P = (val path)
              module S = (val state)

              type path = P.t
              and state = S.t

              module Eval = Eval.Make (P) (S)

              type Ir.builtin +=
                | Gets of Dba.Expr.t * Dba.Expr.t
                | Puts of Dba.Expr.t
                | Exit

              let stdin = "stdin"
              let pointer_size = Kernel_options.Machine.word_size ()

              let cursor =
                Dba.Var.create "stdin_cursor"
                  ~bitsize:(Size.Bit.create pointer_size)
                  ~tag:Temp

              let initialization_callback =
                Some
                  (fun _ state ->
                    let state = S.alloc ~array:stdin state in
                    S.assign cursor
                      (S.Value.constant (Bitvector.zeros pointer_size))
                      state)

              let declaration_callback = None

              let instruction_callback =
                Some
                  (fun (inst : Ast.Instr.t) env : Ir.fallthrough list ->
                    match inst with
                    | Gets (ptr, size) ->
                        [
                          Builtin
                            (Gets
                               ( Script.eval_expr ptr env,
                                 Script.eval_expr size env ));
                        ]
                    | Puts ptr -> [ Builtin (Puts (Script.eval_expr ptr env)) ]
                    | Exit -> [ Builtin Exit ]
                    | _ -> [])

              let process_callback = None

              type io = Read of S.Value.t | Write of S.t * S.Value.t

              let io = P.register_key []

              let incr ptr =
                S.Value.binary Plus ptr
                  (S.Value.constant (Bitvector.ones pointer_size))

              let rec read buf idx size state =
                if size = 0 then S.assign cursor idx state
                else
                  let byte, state =
                    S.select stdin ~addr:idx 1 LittleEndian state
                  in
                  read (incr buf) (incr idx) (size - 1)
                    (S.write ~addr:buf
                       (S.Value.ite
                          (S.Value.binary Eq byte
                             (S.Value.constant (Bitvector.of_char '\000')))
                          (S.Value.constant (Bitvector.of_char '\n'))
                          byte)
                       LittleEndian state)

              let gets buf size _addr path _depth state =
                let buf, state = Eval.safe_eval buf state path in
                let size, state = Eval.safe_eval size state path in
                let size = Bitvector.to_uint (S.get_a_value size state) in
                let idx = S.lookup cursor state in
                P.set io (Read idx :: P.get io path) path;
                let state = read buf idx size state in
                Ok state

              let puts buf _addr path _depth state =
                let buf, state = Eval.safe_eval buf state path in
                P.set io (Write (state, buf) :: P.get io path) path;
                Ok state

              let rec read buf ptr state =
                let byte, state =
                  S.select stdin ~addr:ptr 1 LittleEndian state
                in
                let byte = Bitvector.to_char (S.get_a_value byte state) in
                if byte <> '\000' && byte <> '\n' then (
                  Buffer.add_char buf byte;
                  read buf (incr ptr) state)

              let rec write buf ptr at cur =
                let byte, at = S.read ~addr:ptr 1 LittleEndian at in
                let byte = Bitvector.to_char (S.get_a_value byte cur) in
                if byte <> '\000' then (
                  Buffer.add_char buf byte;
                  write buf (incr ptr) at cur)

              let pp_ios ppf (state, ios) =
                let buf = Buffer.create 128 in
                List.iter
                  (fun io ->
                    Buffer.clear buf;
                    Format.pp_print_newline ppf ();
                    (match io with
                    | Read ptr -> read buf ptr state
                    | Write (at, ptr) -> write buf ptr at state);
                    Format.pp_print_string ppf (Buffer.contents buf))
                  (List.rev ios)

              let tests = Basic_types.String.Htbl.create 64

              let record_io ios state =
                let input = Buffer.create 128 and output = Buffer.create 128 in
                List.iter
                  (fun io ->
                    match io with
                    | Read ptr ->
                        read input ptr state;
                        Buffer.add_char input '\n'
                    | Write (at, ptr) ->
                        write output ptr at state;
                        Buffer.add_char output '\n')
                  (List.rev ios);
                Buffer.truncate input (max 0 (Buffer.length input - 1));
                Buffer.truncate output (max 0 (Buffer.length output - 1));
                Basic_types.String.Htbl.replace tests (Buffer.contents output)
                  (Buffer.contents input)

              let exit _addr path _depth state =
                Logger.result "%a" pp_ios (state, P.get io path);
                record_io (P.get io path) state;
                Error Cut

              let builtin_callback =
                Some
                  (function
                  | Gets (buf, size) -> Some (gets buf size)
                  | Puts buf -> Some (puts buf)
                  | Exit -> Some exit
                  | _ -> None)

              let builtin_printer =
                Some
                  (fun ppf -> function
                    | Gets (buf, size) ->
                        Format.fprintf ppf "gets(%a, %a)"
                          Dba_printer.Ascii.pp_bl_term buf
                          Dba_printer.Ascii.pp_bl_term size;
                        true
                    | Puts buf ->
                        Format.fprintf ppf "puts(%a)"
                          Dba_printer.Ascii.pp_bl_term buf;
                        true
                    | Exit ->
                        Format.pp_print_string ppf "exit()";
                        true
                    | _ -> false)

              let mk_cram_test fmt cmd i input output =
                let inputs = String.split_on_char '\n' input
                and outputs = String.split_on_char '\n' output in
                Format.fprintf fmt
                  "@[<v>@[<v 2>Cram test n°%d@ $ TMP=$(mktemp)%a@ $ cat $TMP | \
                   %s%a@]@ @]"
                  i
                  (fun ppf inputs ->
                    List.iter
                      (fun line -> Format.fprintf ppf "@ $ echo '%s' >> $TMP" line)
                      inputs)
                  inputs cmd
                  (fun ppf outputs ->
                    List.iter
                      (fun line -> Format.fprintf ppf "@ %s" line)
                      outputs)
                  outputs

              let at_exit_callback =
                Option.map
                  (fun path ->
                    let cmd = "./" ^ Kernel_options.ExecFile.get () in
                    fun () ->
                      let oc = open_out path in
                      let fmt = Format.formatter_of_out_channel oc in
                      let n = ref 0 in
                      Basic_types.String.Htbl.iter
                        (fun output input ->
                          Stdlib.incr n;
                          mk_cram_test fmt cmd !n input output)
                        tests;
                      Format.pp_print_flush fmt ();
                      close_out oc)
                  (Output.get_opt ())
            end)
        else None
    end : Exec.PLUGIN)